<?php
/*
Purpose: Basic backend DNS Traversal script.
Author: Mike Jones (mike@mikejones.in)

I started off trying to make it neat and tidy and standards compliant but, well this was what I ended up with!
I wouldn't normally release something with as many hacks and workarounds and known problems as this but I have
had several requests for a copy of the script, so here you go...

Usage: DoLookup($hostname, $type, $callback)

Callback parameter: function ($server, $data) where $server is the server and $data is the response in array format.
*/

if(!isset($GLOBALS['IPv4']))
	$GLOBALS['IPv4']=defined('AF_INET');
if(!isset($GLOBALS['IPv6']))
	$GLOBALS['IPv6']=defined('AF_INET6');
require_once 'Net/DNS.php';

function Cleanup($object)
{
	$arr = Array();
	foreach ($object as $key=>$value)
	{
		if (is_object($value) || is_array($value))
		{
			$arr[$key] = Cleanup($value);
		}
		else
		{
			$arr[$key] = $value;
		}
	}
	if (isset($arr['rdata']))
		unset($arr['rdata']);
	if (isset($arr['rdlength']))
		unset($arr['rdlength']);
	return $arr;
}
function FindRoots()
{
	// Net::DNS is broken for queries to the root, use built in PHP function
	$roots = Array();
	foreach (dns_get_record('.', DNS_NS) as $record)
	{
		$roots[] = $record['target'];
	}
	sort($roots);
	return $roots;
}

function DNSLookup($hostname)
{
	global $recursiveresolver, $recursivecache;
	if (isset($recursivecache[$hostname]))
		return $recursivecache[$hostname];
	if (!$recursivecache)
	{
		$recursivecache = Array();
	}
	$result = Array();
	if (!$recursiveresolver)
	{
		$recursiveresolver = new Net_DNS_Resolver();
		$recursiveresolver->retrans = 1;
		$recursiveresolver->retry = 3;
		$recursiveresolver->recurse = 1;
		$recursiveresolver->tcp_timeout = 3;
	}
	if ($GLOBALS['IPv4'])//defined('AF_INET'))
	{
		$response = $recursiveresolver->query($hostname, 'A', 'IN');
		if ($response && $response->answer)
		{
			foreach ($response->answer as $answer)
			{
				if ($answer->type == 'A' && $answer->address)
				{
					$result[] = $answer->address;
				}
			}
		}
	}
	if ($GLOBALS['IPv6'])//defined('AF_INET6'))
	{
		$response = $recursiveresolver->query($hostname, 'AAAA', 'IN');
		if ($response && $response->answer)
		{
			foreach ($response->answer as $answer)
			{
				if ($answer->type == 'AAAA' && $answer->address)
				{
					$result[] = '[' . $answer->address . ']';
				}
			}
		}
	}
	$recursivecache[$hostname] = $result;
	return $result;
}

function SendQuery($hostname, $type, $server)
{
	global $resolver;
	if (!$resolver)
	{
		$resolver = new Net_DNS_Resolver();
	}
	$resolver->nameservers = Array($server);
	$resolver->useEnhancedSockets = false; // a lot easier than dealing with the totally different IPv6 behaviour depending on library used!
	$resolver->retrans = 1;
	$resolver->retry = 3;
	$resolver->recurse = 0;
	$resolver->tcp_timeout = 3;
	$response = $resolver->rawQuery($hostname, $type, 'IN');
	if ($response) {
		return Array('server'=>$response->answerfrom,'header'=>Cleanup($response->header), 'answer'=>Cleanup($response->answer), 'authority'=>Cleanup($response->authority), 'additional'=>Cleanup($response->additional));
	}
	return Array('header'=>false, 'answer'=>Array(), 'authority'=>Array(), 'additional'=>Array());
}

function DoLookup($hostname, $type, $callback)
{
	$replies = Array();
	$fakedelegation = Array();
	foreach (FindRoots() as $rootserver)
	{
		$addresses = DNSLookup($rootserver);
		foreach($addresses as $addr)
		{
			$replies[$addr] = false;
			$fakedelegation[] = Array('address'=>$addr, 'name'=>$rootserver);
		}
	}
	$callback("IANA Root", Array('delegations'=>$fakedelegation, 'header'=>Array('rcode'=>'NOERROR'), 'answer'=>Array(), 'authority'=>Array(), 'additional'=>Array()));
	
	$done = false;
	while(!$done)
	{
		$more = false;
		foreach($replies as $server=>$data)
		{
			if ($replies[$server] === false)
			{
				$replies[$server] = SendQuery($hostname, $type, $server);
				if ($replies[$server]['header']['aa'] == 0 && $replies[$server]['header']['rcode'] == 'NOERROR' && count($replies[$server]['authority']) > 0)
				{
					$replies[$server]['delegations'] = Array();
					foreach ($replies[$server]['authority'] as $referral)
					{
						foreach($replies[$server]['additional'] as $glue)
						{
							if (strtolower($glue['name']) == strtolower($referral['nsdname']) && $glue['type'] == 'A' && $GLOBALS['IPv4'])
							{
								$address = $glue['address'];
								if (!isset($replies[$server]['delegations'][md5($referral['nsdname'].$address)]))
								{
									$replies[$server]['delegations'][md5($referral['nsdname'].$address)] = Array('address'=>$address, 'name'=>$referral['nsdname']);
									$replies[$address] = false;
									$more = true;
								}
								$replies[$server]['delegations'][md5($referral['nsdname'].$address)]['glue'] = true;
							}
							if (strtolower($glue['name']) == strtolower($referral['nsdname']) && $glue['type'] == 'AAAA' && $GLOBALS['IPv6'])
							{
								$address = '[' . $glue['address'] . ']';
								if (!isset($replies[$server]['delegations'][md5($referral['nsdname'].$address)]))
								{
									$replies[$server]['delegations'][md5($referral['nsdname'].$address)] = Array('address'=>$address, 'name'=>$referral['nsdname']);
									$replies[$address] = false;
									$more = true;
								}
								$replies[$server]['delegations'][md5($referral['nsdname'].$address)]['glue'] = true;
							}
						}
						$delegations = DNSLookup($referral['nsdname']);
						foreach ($delegations as $address)
						{
							if (!isset($replies[$server]['delegations'][md5($referral['nsdname'].$address)]))
							{
								$replies[$server]['delegations'][md5($referral['nsdname'].$address)] = Array('address'=>$address, 'name'=>$referral['nsdname']);
								$replies[$address] = false;
								$more = true;
							}
							$replies[$server]['delegations'][md5($referral['nsdname'].$address)]['lookup'] = true;
						}
					}
				}
				$callback($server, $replies[$server]);
			}
		}
		$done = !$more;
	}
	return $replies;
}
?>
